/*
	(C) 1995-2001 AROS - The Amiga Research OS
	$Id: openwindow.c,v 1.34 2021/05/15 02:15:37 jacadcaps Exp $

	Desc: Intuition function OpenWindow()
	Lang: english
*/
#include <exec/memory.h>
#include <graphics/layers.h>
#include <graphics/gfx.h>
#include <intuition/intuition.h>
#include <intuition/imageclass.h>
#include <utility/tagitem.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/utility.h>
#include <proto/intuition.h>
#include <proto/layers.h>
#include <exec/ports.h>
#include <hardware/atomic.h>
#include "intuition_intern.h"
#include "inputhandler.h"
#include "inputhandler_actions.h"
#include "intuition_extend.h"
#include "boopsigadgets.h"
#include "transplayers.h"
#include "mosmisc.h"
#include "inputhandlermos.h"
#include "intuition_customizesupport.h"
#include "layermanager.h"

#ifndef DEBUG_OpenWindow
#   define DEBUG_OpenWindow 0
#endif
#undef DEBUG
#define DEBUG 0
#if DEBUG_OpenWindow
#   define DEBUG 1
#endif
#	include <aros/debug.h>

struct OpenWindowActionMsg
{
	struct IntuiActionMsg msg;
	struct Window *window;
	struct BitMap *bitmap;
	struct Hook *backfillhook;
	struct Region *shape;
	struct Layer *parentlayer;
	BOOL visible;
	BOOL success;
	BOOL realtime;
};

static VOID int_openwindow(struct OpenWindowActionMsg *msg,
						   struct IntuitionBase *IntuitionBase);

/*****************************************************************************

	NAME */

AROS_LH1(struct Window *, OpenWindow,

		 /*  SYNOPSIS */
		 AROS_LHA(struct NewWindow *, newWindow, A0),

		 /*  LOCATION */
		 struct IntuitionBase *, IntuitionBase, 34, Intuition)

/*  FUNCTION
	Opens a new window with the characteristics specified in
	newWindow.

	INPUTS
	newWindow - How you would like your new window.

	RESULT
	A pointer to the new window or NULL if it couldn't be
	opened. Reasons for this might be lack of memory or illegal
	attributes.

	NOTES

	EXAMPLE

	BUGS

	SEE ALSO
	CloseWindow(), ModifyIDCMP()

	INTERNALS

	HISTORY
	29-10-95    digulla automatically created from
			    intuition_lib.fd and clib/intuition_protos.h

*****************************************************************************/
{
	AROS_LIBFUNC_INIT
	AROS_LIBBASE_EXT_DECL(struct IntuitionBase *,IntuitionBase)

	struct OpenWindowActionMsg msg;
	struct NewWindow	nw;
	struct Window	*w = NULL, *helpgroupwindow = NULL, *parentwin = NULL;
	struct TagItem	*tag, *tagList;
	struct RastPort	*rp;
	struct Hook		*backfillhook = LAYERS_BACKFILL;
	struct Region	*shape = NULL;
	struct IBox		*zoombox = NULL;
	struct Image	*AmigaKey = NULL;
	struct Image	*Checkmark = NULL;
	struct Layer	*parentl = NULL;
#ifdef SKINS
	struct SkinInfo	*skininfo = NULL;
	BOOL hasskininfo = FALSE;
	struct Region   *usertranspregion = NULL;
	struct Hook     *usertransphook = NULL;
	ULONG            useropacity = 0xFFFFFFFF;
	ULONG            usersourcealpha = FALSE;
	ULONG            userhitlevel = 0;
	ULONG            usereffectbits = 0xFFFFFFFF;
	struct MsgPort  *userport = NULL;
	ULONG *app = NULL;
	ULONG            AspectSizeNumerator = 1;
	ULONG            AspectSizeDenominator = 1;
	ULONG            AspectSizeExtraWidth = 0;
	ULONG            AspectSizeExtraHeight = 0;
	BOOL             uselayermanagerset = FALSE;
	BOOL             realtime = FALSE;
#endif

	STRPTR		pubScreenName = NULL;
	UBYTE		*screenTitle = NULL;
	BOOL		autoAdjust = FALSE, pubScreenFallBack = FALSE;
	ULONG		innerWidth = ~0L;
	ULONG		innerHeight = ~0L;
	WORD		mousequeue = DEFAULTMOUSEQUEUE;
	WORD		repeatqueue = 3; /* stegerg: test on my Amiga suggests this */
	ULONG		moreFlags = 0;
	ULONG		helpgroup = 0;
	ULONG		extrabuttons = 0, extrabuttonsid = ETI_Dummy;

	//    ULONG		lock;
	ULONG        windowvisible = TRUE;
	BOOL		 driver_init_done = FALSE, have_helpgroup = FALSE;
	BOOL         do_setwindowpointer = FALSE;
	ULONG		 thinborders = 0;

	ASSERT_VALID_PTR_ROMOK(newWindow);

	D(bug("OpenWindow (%p = { Left=%d Top=%d Width=%d Height=%d })\n"
	      , newWindow
	      , newWindow->LeftEdge
	      , newWindow->TopEdge
	      , newWindow->Width
	      , newWindow->Height
	     ));

	nw = *newWindow;

#define WFLG_PRIVATEFLAGS (WFLG_WINDOWREFRESH |\
WFLG_WINDOWTICKED  | WFLG_VISITOR       | \
WFLG_ZOOMED        |\
WFLG_WINDOWACTIVE )
	/* jDc*/
	/* WFLG_WBENCHWINDOW  | \*/
	/* this is used by WORKBENCH! */
	/* WFLG_HASZOOM       | \*/
	/* do NOT filter this! how do you think apps could manage to get a zoom image with struct NewWindow? */

	nw.Flags &= ~WFLG_PRIVATEFLAGS;

#ifdef __MORPHOS__
	if (IsDuringSelfResize(IntuitionBase))
	{
		dprintf("Major bug: calling OpenWindow() from within MUIM_Draw task %lx\n",(ULONG)FindTask(0));
		return NULL;
	}

#endif
	if (newWindow->Flags & WFLG_NW_EXTENDED)
	{
		tagList = ((struct ExtNewWindow *)newWindow)->Extension;
#if 1
		/* Sanitycheck the taglist pointer. Some Am*gaOS 1.3/2.x era
		 * apps have WFLG_NW_EXTENDED set with bogus Extension taglist
		 * pointer... (older CygnusED for example) - Piru
		 */
		if (((ULONG) tagList & 1) || !TypeOfMem(tagList))
		{
			tagList = NULL;
		}
#endif
	}
	else
	{
		tagList = NULL;
	}

	DEBUG_OPENWINDOW(dprintf("OpenWindow: NewWindow 0x%lx TagList 0x%lx\n",
				 newWindow, tagList));

	if (tagList)
	{
		ASSERT_VALID_PTR_ROMOK(tagList);

		/* Look at WA_Flags first, since boolean tags override part of it
		 * even if they appear before it.
		 */
		nw.Flags |= (GetTagData(WA_Flags, nw.Flags, tagList) & ~WFLG_PRIVATEFLAGS);

		while ((tag = NextTagItem ((struct TagItem **)&tagList)))
		{
			/* ASSERT_VALID_PTR_ROMOK(tag); */

			DEBUG_OPENWINDOW(dprintf("OpenWindow: Tag 0x%08lx 0x%08lx\n",
						 tag->ti_Tag, tag->ti_Data));

			switch (tag->ti_Tag)
			{
			case WA_Left:
				nw.LeftEdge	= tag->ti_Data;
				break;
			case WA_Top:
				nw.TopEdge	= tag->ti_Data;
				break;
			case WA_Width:
				if ((tag->ti_Data != ~0) && (tag->ti_Data > 32767))
				{
					DEBUG_OPENWINDOW(dprintf("OpenWindow: WA_Width %ld too large/negative !!!\n", tag->ti_Data));

					goto failexit;
				}
				nw.Width	= tag->ti_Data;
				break;
			case WA_Height:
				if ((tag->ti_Data != ~0) && (tag->ti_Data > 32767))
				{
					DEBUG_OPENWINDOW(dprintf("OpenWindow: WA_Height %ld too large/negative !!!\n", tag->ti_Data));

					goto failexit;
				}
				nw.Height	= tag->ti_Data;
				break;
			case WA_IDCMP:
				nw.IDCMPFlags   = tag->ti_Data;
				break;
			case WA_MinWidth:
				nw.MinWidth     = tag->ti_Data;
				break;
			case WA_MinHeight:
				nw.MinHeight    = tag->ti_Data;
				break;
			case WA_MaxWidth:
				nw.MaxWidth     = tag->ti_Data;
				break;
			case WA_MaxHeight:
				nw.MaxHeight    = tag->ti_Data;
				break;

			case WA_Gadgets:
				nw.FirstGadget  = (struct Gadget *)(tag->ti_Data);
				break;
			case WA_Title:
				nw.Title	= (UBYTE *)(tag->ti_Data);
				break;

			case WA_ScreenTitle:
				screenTitle = (UBYTE *)tag->ti_Data;
				break;
			case WA_AutoAdjust:
				autoAdjust  = (tag->ti_Data != 0);
				break;
			case WA_InnerWidth:
				innerWidth  = tag->ti_Data;
				break;
			case WA_InnerHeight:
				innerHeight = tag->ti_Data;
				break;

#define MODIFY_FLAG(name)   if (tag->ti_Data) \
nw.Flags |= (name); else nw.Flags &= ~(name)
#define MODIFY_MFLAG(name)  if (tag->ti_Data) \
moreFlags |= (name); else moreFlags &= ~(name)

			case WA_SizeGadget:
				MODIFY_FLAG(WFLG_SIZEGADGET);
				break;
			case WA_DragBar:
				MODIFY_FLAG(WFLG_DRAGBAR);
				break;
			case WA_DepthGadget:
				MODIFY_FLAG(WFLG_DEPTHGADGET);
				break;
			case WA_CloseGadget:
				MODIFY_FLAG(WFLG_CLOSEGADGET);
				break;
			case WA_Backdrop:
				MODIFY_FLAG(WFLG_BACKDROP);
				break;
			case WA_ReportMouse:
				MODIFY_FLAG(WFLG_REPORTMOUSE);
				break;
			case WA_NoCareRefresh:
				MODIFY_FLAG(WFLG_NOCAREREFRESH);
				break;
			case WA_Borderless:
				MODIFY_FLAG(WFLG_BORDERLESS);
				break;
			case WA_Activate:
				MODIFY_FLAG(WFLG_ACTIVATE);
				break;
			case WA_RMBTrap:
				MODIFY_FLAG(WFLG_RMBTRAP);
				break;
			case WA_WBenchWindow:
				MODIFY_FLAG(WFLG_WBENCHWINDOW);
				break;
			case WA_SizeBRight:
				MODIFY_FLAG(WFLG_SIZEBRIGHT);
				break;
			case WA_SizeBBottom:
				MODIFY_FLAG(WFLG_SIZEBBOTTOM);
				break;
			case WA_GimmeZeroZero:
				MODIFY_FLAG(WFLG_GIMMEZEROZERO);
				break;
			case WA_NewLookMenus:
				MODIFY_FLAG(WFLG_NEWLOOKMENUS);
				break;
			case WA_Zoom:
				zoombox = (struct IBox *)tag->ti_Data;
				DEBUG_OPENWINDOW(dprintf("OpenWindow: zoom %d %d %d %d\n",
							 zoombox->Left, zoombox->Top, zoombox->Width, zoombox->Height));
				MODIFY_FLAG(WFLG_HASZOOM);
				break;

			case WA_DetailPen:
				if (nw.DetailPen == 0xFF)
					nw.DetailPen = tag->ti_Data;
				break;

			case WA_BlockPen:
				if (nw.BlockPen == 0xFF)
					nw.BlockPen = tag->ti_Data;
				break;

			case WA_CustomScreen:
				nw.Screen = (struct Screen *)(tag->ti_Data);
				nw.Type = CUSTOMSCREEN;
				break;

			case WA_SuperBitMap:
				nw.Flags |= WFLG_SUPER_BITMAP;
				nw.BitMap = (struct BitMap *)(tag->ti_Data);
				break;

			case WA_SimpleRefresh:
				if (tag->ti_Data)
					nw.Flags |= WFLG_SIMPLE_REFRESH;
				break;

			case WA_SmartRefresh:
				if (tag->ti_Data)
					nw.Flags |= WFLG_SMART_REFRESH;
				break;

			case WA_PubScreenFallBack:
				pubScreenFallBack = (tag->ti_Data ? TRUE : FALSE);
				break;

			case WA_PubScreenName:
				//nw.Type = PUBLICSCREEN; //handled below!!
				pubScreenName = (STRPTR)tag->ti_Data;
				break;

			case WA_PubScreen:
				nw.Type = PUBLICSCREEN;
				nw.Screen = (struct Screen *)tag->ti_Data;
				break;

			case WA_BackFill:
				backfillhook = (struct Hook *)tag->ti_Data;
				break;

			case WA_MouseQueue:
				mousequeue = tag->ti_Data;
				break;

				/* These two are not implemented in AmigaOS */
			case WA_WindowName:
			case WA_Colors:
				break;

			case WA_NotifyDepth:
				MODIFY_MFLAG(WMFLG_NOTIFYDEPTH);
				break;

			case WA_RptQueue:
				repeatqueue = tag->ti_Data;
				break;

			case WA_Checkmark:
				Checkmark = (struct Image *)tag->ti_Data;
				break;

			case WA_AmigaKey:
				AmigaKey = (struct Image *)tag->ti_Data;
				break;

			case WA_HelpGroup:
				helpgroup = (ULONG)tag->ti_Data;
				have_helpgroup = TRUE;
				break;

			case WA_HelpGroupWindow:
				helpgroupwindow = (struct Window *)tag->ti_Data;
				break;

			case WA_MenuHelp:
				MODIFY_MFLAG(WMFLG_MENUHELP);
				break;

			case WA_PointerDelay:
				MODIFY_MFLAG(WMFLG_POINTERDELAY);
				break;

			case WA_TabletMessages:
				MODIFY_MFLAG(WMFLG_TABLETMESSAGES);
				break;

			case WA_ExtraTitlebarGadgets:
				extrabuttons = (ULONG)tag->ti_Data;
				break;

			case WA_ExtraGadgetsStartID:
				extrabuttonsid = (ULONG)tag->ti_Data;
				break;

			case WA_ExtraGadget_Iconify:
				if (tag->ti_Data)
				{
					extrabuttons |= ETG_ICONIFY;
				}
				else
				{
					extrabuttons &= ~ETG_ICONIFY;
				};
				break;

			case WA_ExtraGadget_Lock:
				if (tag->ti_Data)
				{
					extrabuttons |= ETG_LOCK;
				}
				else
				{
					extrabuttons &= ~ETG_LOCK;
				};
				break;

			case WA_ExtraGadget_MUI:
				if (tag->ti_Data)
				{
					extrabuttons |= ETG_MUI;
				}
				else
				{
					extrabuttons &= ~ETG_MUI;
				};
				break;

			case WA_ExtraGadget_PopUp:
				if (tag->ti_Data)
				{
					extrabuttons |= ETG_POPUP;
				}
				else
				{
					extrabuttons &= ~ETG_POPUP;
				};
				break;

			case WA_ExtraGadget_Snapshot:
				if (tag->ti_Data)
				{
					extrabuttons |= ETG_SNAPSHOT;
				}
				else
				{
					extrabuttons &= ~ETG_SNAPSHOT;
				};
				break;

			case WA_ExtraGadget_Jump:
				if (tag->ti_Data)
				{
					extrabuttons |= ETG_JUMP;
				}
				else
				{
					extrabuttons &= ~ETG_JUMP;
				};
				break;

#ifdef SKINS
			case WA_SkinInfo:
				skininfo = (struct SkinInfo *)tag->ti_Data;
				hasskininfo = TRUE;
				break;

			case WA_TransparentRegion:
				usertranspregion = (struct Region *)tag->ti_Data;
				usertransphook = NULL; //doesn't make sense
				break;

			case WA_TransparentRegionHook:
				usertransphook = (struct Hook *)tag->ti_Data;
				usertranspregion = NULL;
				break;

			case WA_UserPort:
				userport = (struct MsgPort *)tag->ti_Data;
				break;

			case WA_IAmMUI:
				app = (ULONG*)tag->ti_Data;
				MODIFY_MFLAG(WMFLG_IAMMUI);
				break;

			case WA_SelfResize:
				MODIFY_MFLAG(WMFLG_SELFRESIZING);
				break;

			case WA_IgnoreMe:
				MODIFY_MFLAG(WMFLG_IGNOREME);
				break;

			case WA_ToolbarWindow:
				MODIFY_MFLAG(WMFLG_TOOLBARWINDOW);
				if (tag->ti_Data)
				{
					MODIFY_FLAG(WFLG_BORDERLESS); //only those can be used as toolbar windows!
				}
				break;

			case WA_FrontWindow:
				MODIFY_MFLAG(WMFLG_FRONTWINDOW);
				break;

			case WA_Parent:
				parentwin = (struct Window *)tag->ti_Data;
				break;

			case WA_ThinBorders:
				thinborders = (ULONG)tag->ti_Data;
				moreFlags |= WMFLG_THINBORDERS;
				break;

			case WA_Opacity:
				useropacity = (ULONG)tag->ti_Data;
				break;

			case WA_HasAlpha:
				usersourcealpha = (BOOL)tag->ti_Data;
				break;

			case WA_HitLevel:
				userhitlevel = min(256,max(0,tag->ti_Data));
				break;

			case WA_SizeNumerator:
				AspectSizeNumerator = (ULONG)tag->ti_Data;
				break;

			case WA_SizeDenominator:
				AspectSizeDenominator = (ULONG)tag->ti_Data;
				break;

			case WA_SizeExtraWidth:
				AspectSizeExtraWidth = (ULONG)tag->ti_Data;
				break;
				
			case WA_SizeExtraHeight:
				AspectSizeExtraHeight = (ULONG)tag->ti_Data;
				break;

			case WA_ShadowTop:
				if (usereffectbits == 0xFFFFFFFF) usereffectbits = 0;
				usereffectbits |= 0x7;
				break;

			case WA_ShadowLeft:
				if (usereffectbits == 0xFFFFFFFF) usereffectbits = 0;
				usereffectbits |= 0x38;
				break;

			case WA_ShadowRight:
				if (usereffectbits == 0xFFFFFFFF) usereffectbits = 0;
				usereffectbits |= 0x1C0;
				break;

			case WA_ShadowBottom:
				if (usereffectbits == 0xFFFFFFFF) usereffectbits = 0;
				usereffectbits |= 0xE00;
				break;

			case WA_VisibleOnMaximize:
				MODIFY_MFLAG(WMFLG_NOHIDEONZOOM);
				break;
			
			case WA_Unicode:
				if (tag->ti_Data)
					moreFlags |= WMFLG_UNICODE;
				else
					moreFlags &= ~WMFLG_UNICODE;
				break;
				
			case WA_RealtimeLayer:
				realtime = tag->ti_Data;
				break;
#endif

#ifndef __MORPHOS__
			case WA_Shape:
				shape = (struct Region *)tag->ti_Data;
				break;

			case WA_Parent:
				parentwin = ((struct Window *)tag->ti_Data);
				parentl   = parentwin->WLayer;
				break;

			case WA_Visible:
				windowvisible = (ULONG)tag->ti_Data;
				break;

#endif

			case WA_Pointer:
			case WA_BusyPointer:
#ifdef SKINS
			case WA_PointerType:
#endif
				do_setwindowpointer = TRUE;
				break;

			case WA_UseLayerManager:
				uselayermanagerset = TRUE;
				if (tag->ti_Data)
					moreFlags |= WMFLG_LAYERMANAGER;
				else
					moreFlags &= ~WMFLG_LAYERMANAGER;
				break;
			
			} /* switch Tag */

		} /* while ((tag = NextTagItem (&tagList))) */

	} /* if (tagList) */

	if (nw.Flags & WFLG_SIZEGADGET)
	{
		if (!(nw.Flags & (WFLG_SIZEBRIGHT | WFLG_SIZEBBOTTOM)))
		{
			nw.Flags |= WFLG_SIZEBRIGHT;
		}

		//jDc: tested behavior of intuition68k
		nw.Flags |= WFLG_HASZOOM;
	}
	else
	{
		nw.Flags &= ~(WFLG_SIZEBRIGHT|WFLG_SIZEBBOTTOM);
	}

	if (nw.Flags & WFLG_BORDERLESS)
	{
		nw.Flags &= ~(WFLG_SIZEBRIGHT|WFLG_SIZEBBOTTOM|WFLG_SIZEGADGET);
	}

	/* Find out on which Screen the window must open */

		/* (cyfm 03/03/03 check for nw.Type == PUBLICSCREEN as well, some programs
		 *  like TurboPrint GraphicPublisher specify {WA_PubScreen, NULL} and want
		 *  to open on the default public screen that way 
		 */

	#ifdef SKINS
	if (!ScreenValid(nw.Screen,IntuitionBase)) nw.Screen = NULL;
	#endif

	if (!nw.Screen && (pubScreenName || (nw.Type == PUBLICSCREEN)))
	{
		struct Screen *pubs = 0;
		moreFlags |= WMFLG_DO_UNLOCKPUBSCREEN;
		pubs = LockPubScreen(pubScreenName);
		if (!pubs && pubScreenFallBack)
		{
			nw.Screen = LockPubScreen(NULL);
		}
		if (pubs)
		{
			nw.Screen = pubs;
		}
		nw.Type = PUBLICSCREEN;
		if (nw.Screen) nw.Flags |= WFLG_VISITOR;
	}

	if (!nw.Screen && nw.Type == WBENCHSCREEN)
	{
		nw.Screen = LockPubScreen("Workbench");
		if (nw.Screen)
		{
			moreFlags |= WMFLG_DO_UNLOCKPUBSCREEN;
			if (nw.Screen) nw.Flags |= WFLG_VISITOR;
		}
	}

	if (nw.Screen == NULL)
		goto failexit;

#ifdef __MORPHOS__
	w = NewObjectA(IBase->windowclass,0,0);
#else
	w  = AllocMem (sizeof(struct IntWindow), MEMF_CLEAR);
#endif

	DEBUG_OPENWINDOW(dprintf("OpenWindow: Window 0x%lx\n", w));

	/* nlorentz: For now, creating a rastport becomes the responsibility of
	   intui_OpenWindow(). This is because intui_OpenWindow() in
	   config/hidd/intuition_driver.c must call CreateUpfrontLayer(),
	   and that will create a rastport for the layer/window, and we don't
	   want two rastports pr. window.
	   Alternatively we may create a layers_driver.c driver for layers,
	   and then call CreateUpfrontLayer() here from openwindow.
	   For the Amiga window<-->X11 window stuff, the layers driver
	   would just allocate a layer struct, a rastport and
	   put the rasport into layer->RastPort, so we
	   could get it inside this routine and put it into
	   window->RPort;.

	*/

	if (NULL == w)
		goto failexit;

	DEBUG_OPENWINDOW(dprintf("OpenWindow: Flags 0x%lx MoreFlags 0x%lx IDCMP 0x%lx\n",
				nw.Flags, moreFlags, nw.IDCMPFlags));

	w->WScreen = nw.Screen;

	#ifdef SKINS
	if (moreFlags & WMFLG_IAMMUI)
		w->Flags |= WFLG_SIMPLE_REFRESH;
	
	if (userport)
	{
		w->UserPort = userport;
		((struct IntWindow *)w)->specialflags |= SPFLAG_USERPORT;
	}
	#endif

	if (IS(w->WScreen)->Compositing)
	{
		if (!uselayermanagerset)
		{
			if (0 == (nw.Flags & WFLG_BORDERLESS))
				moreFlags |= WMFLG_LAYERMANAGER;
		}

		if (realtime && 0 == (nw.Flags & WFLG_BORDERLESS))
			goto failexit;
	}
	else
	{
		moreFlags &= ~WMFLG_LAYERMANAGER;
		if (realtime)
			goto failexit;
	}

	if (!ModifyIDCMP (w, nw.IDCMPFlags))
		goto failexit;

	((struct IntWindow *)w)->extrabuttons = extrabuttons;
	((struct IntWindow *)w)->extrabuttonsid = extrabuttonsid;

	/*    w->RPort	   = rp; */

	//w->FirstGadget = nw.FirstGadget;

	w->DetailPen = (nw.DetailPen != 0xFF) ? nw.DetailPen : w->WScreen->DetailPen;
	w->BlockPen  = (nw.BlockPen != 0xFF)  ? nw.BlockPen  : w->WScreen->BlockPen;

	/* Copy flags */
	w->Flags = nw.Flags;
	w->MoreFlags = moreFlags;

	if (!(w->Flags & WFLG_BORDERLESS))
	{
#ifdef SKINS
		if (hasskininfo)
		{
			struct SkinInfo *skininfo = 0;

			skininfo = &IS(w->WScreen)->SkinInfo;

			w->BorderLeft = skininfo->BorderLeft;
			w->BorderTop = skininfo->BorderTop;
			w->BorderBottom = skininfo->BorderBottom;
			w->BorderRight = skininfo->BorderRight;
		} else {
#endif
			w->BorderLeft   = w->WScreen->WBorLeft;
			w->BorderRight  = w->WScreen->WBorRight;
			w->BorderTop    = w->WScreen->WBorBottom;
			w->BorderBottom = w->WScreen->WBorBottom;
#ifdef SKINS
		}
#endif
	}

	if (nw.Title || (w->Flags & (WFLG_DRAGBAR | WFLG_CLOSEGADGET | WFLG_DEPTHGADGET)))
	{
		/* this is a hack. the correct way to "correct" (increase if necessary)
		   the w->Border??? items would be to check all GACT_???BORDER gadgets
		   (inclusive sysgadgets which are GACT_????BORDER gadgets as well) in
		   nw.FirstGadget (or WA_Gadgets tag) and all sysgadgets and then
		   make sure that each window border is big enough so that none of these
		   gadgets extends outside the window border area */

#ifdef SKINS
		if (hasskininfo)
		{
			struct SkinInfo *skininfo = 0;

			skininfo = &IS(w->WScreen)->SkinInfo;

			w->BorderTop = skininfo->BorderTopTitle;
		} else {
#endif
			w->BorderTop    = w->WScreen->WBorTop;

			/* Georg Steger: ??? font ??? */
			if (w->WScreen->Font)
				w->BorderTop += ((struct IntScreen *)(w->WScreen))->DInfo.dri_Font->tf_YSize + 1;
			else
				w->BorderTop += GfxBase->DefaultFont->tf_YSize + 1;
#ifdef SKINS
		}
#endif
	}

	/* look for GACT_???BORDER gadgets which increase the BorderSizes */

	if (nw.FirstGadget)
	{
		struct Gadget *gad;

		for(gad = nw.FirstGadget; gad; gad = gad->NextGadget)
		{
			WORD gadx1, gady1, gadx2, gady2;

			if (gad->Activation & GACT_LEFTBORDER)
			{
				/* may never be GFLG_RELRIGHT / GFLG_RELWIDTH */

				gadx2 = gad->LeftEdge + gad->Width - 1;
				if (gadx2 >= w->BorderLeft) w->BorderLeft = gadx2/* + 1*/;
			}

			if (gad->Activation & GACT_TOPBORDER)
			{
				/* may never be GFLG_RELBOTTOM / GFLG_RELHEIGHT */

				gady2 = gad->TopEdge + gad->Height - 1;
				if (gady2 >= w->BorderTop) w->BorderTop = gady2/* + 1*/;
			}

			if (gad->Activation & GACT_RIGHTBORDER)
			{
				/* must be GFLG_RELRIGHT but never GFLG_RELWIDTH */

				gadx1 = -gad->LeftEdge;
				if (gadx1 >= w->BorderRight) w->BorderRight = gadx1/* + 1*/;
			}

			if (gad->Activation & GACT_BOTTOMBORDER)
			{
				/* must be GFLG_RELBOTTOM but never GFLG_RELHEIGHT */

				gady1 = -gad->TopEdge;
				if (gady1 >= w->BorderBottom) w->BorderBottom = gady1/* + 1*/;
			}

		} /* for(gad = nw.FirstGadget; gad; gad = gad->NextGadget) */

	} /* if (nw.FirstGadget) */

#ifdef SKINS
	if (hasskininfo)
	{
		struct SkinInfo *skininfo = 0;

		skininfo = &IS(w->WScreen)->SkinInfo;

		if (!(w->Flags & WFLG_SIZEBRIGHT)) if (w->BorderRight > skininfo->BorderRight)
		{
			w->Flags |= WFLG_SIZEBRIGHT;
		}

		if (!(w->Flags & WFLG_SIZEBBOTTOM)) if (w->BorderBottom > skininfo->BorderBottom)
		{
			w->Flags |= WFLG_SIZEBBOTTOM;
		}
	} else {
#endif
		if (!(w->Flags & WFLG_SIZEBRIGHT)) if (w->BorderRight > w->WScreen->WBorRight)
		{
			w->Flags |= WFLG_SIZEBRIGHT;
		}

		if (!(w->Flags & WFLG_SIZEBBOTTOM)) if (w->BorderBottom > w->WScreen->WBorBottom)
		{
			w->Flags |= WFLG_SIZEBBOTTOM;
		}
#ifdef SKINS
	}
#endif
	
	//    if ((w->Flags & WFLG_SIZEGADGET) &&
	//       (w->Flags & (WFLG_SIZEBRIGHT | WFLG_SIZEBBOTTOM)))
	{
		IPTR sizewidth = 16, sizeheight = 16;
		struct Image *im;
		struct DrawInfo *dri;

		if ((dri = GetScreenDrawInfo(w->WScreen)))
		{
			struct TagItem imtags[] =
			    {
				    {SYSIA_DrawInfo, (STACKIPTR)dri},
				    {SYSIA_Which, SIZEIMAGE},
				    {SYSIA_Size, w->WScreen->Flags & SCREENHIRES ? SYSISIZE_MEDRES : SYSISIZE_LOWRES},
				    {TAG_DONE,}
			    };

			if ((im = NewObjectA(NULL, SYSICLASS, imtags)))
			{
				GetAttr(IA_Width, (Object *)im, &sizewidth);
				GetAttr(IA_Height, (Object *)im, &sizeheight);

				DisposeObject(im);
			}
			FreeScreenDrawInfo(w->WScreen, dri);
		}

		if (w->Flags & WFLG_SIZEBRIGHT)
		{
			if (w->BorderRight < sizewidth) w->BorderRight = sizewidth;
		}

		if (w->Flags & WFLG_SIZEBBOTTOM)
		{
			if (w->BorderBottom < sizeheight) w->BorderBottom = sizeheight;
		}

		IW(w)->sizeimage_width = sizewidth;
		IW(w)->sizeimage_height = sizeheight;

		/* now increase window size if it's necessary */
	}

#ifdef SKINS
	IW(w)->custombackfill.h_Entry = (HOOKFUNC)HookEntry;
	IW(w)->custombackfill.h_SubEntry = (HOOKFUNC)GradientizeBackfillFunc;
	IW(w)->custombackfill.h_Data = &IW(w)->hd;
	IW(w)->hd.intuitionBase = IntuitionBase;
		
	IW(w)->usertranspregion = usertranspregion;
	IW(w)->usertransphook = usertransphook;
	IW(w)->useropacity = useropacity;
	IW(w)->usersourcealpha = usersourcealpha;
	IW(w)->userhitlevel = userhitlevel;

	IW(w)->AspectSizeNumerator = AspectSizeNumerator;
	IW(w)->AspectSizeDenominator = AspectSizeDenominator;
	IW(w)->AspectSizeExtraWidth = AspectSizeExtraWidth;
	IW(w)->AspectSizeExtraHeight = AspectSizeExtraHeight;

	if (usereffectbits == 0xFFFFFFFF)
	{
		IW(w)->EffectBits = (w->Flags & WFLG_BORDERLESS) ? 0 : 0xFFF;
		if (usertranspregion || usertransphook) IW(w)->EffectBits = 0;
		IW(w)->DefaultEffectBits = TRUE;
	}
	else
	{
		IW(w)->EffectBits = usereffectbits;
		IW(w)->DefaultEffectBits = FALSE;
	}
#endif

	/* Now verifies that the inner#? values are within valid range. - Piru */
	if (innerWidth != ~0L)
	{
		if ((LONG) innerWidth < 0) goto failexit;
		innerWidth += w->BorderLeft + w->BorderRight;
		if (innerWidth > 32767) goto failexit;

		nw.Width = innerWidth;
	}
	if (innerHeight != ~0L)
	{
		if ((LONG) innerHeight < 0) goto failexit;
		innerHeight += w->BorderTop + w->BorderBottom;
		if (innerHeight > 32767) goto failexit;

		nw.Height = innerHeight;
	}

	{
		LONG parentwidth;
		LONG parentheight;

		parentwidth  = parentwin ? parentwin->Width : w->WScreen->Width;
		parentheight = parentwin ? parentwin->Height : w->WScreen->Height;

		w->Width   = (nw.Width  != ~0) ? nw.Width  : parentwidth - nw.LeftEdge;
		w->Height  = (nw.Height != ~0) ? nw.Height : parentheight - nw.TopEdge;

		if (autoAdjust)
		{

			if (w->Width  > parentwidth)  w->Width  = parentwidth;
			if (w->Height > parentheight) w->Height = parentheight;

			if (nw.LeftEdge < 0) nw.LeftEdge = 0;
			if (nw.TopEdge < 0) nw.TopEdge = 0;

			if ((nw.LeftEdge + w->Width) > parentwidth)
				nw.LeftEdge = parentwidth - w->Width;
			if ((nw.TopEdge + w->Height) > parentheight)
				nw.TopEdge = parentheight - w->Height;
		}

		w->GZZWidth  = w->Width  - w->BorderLeft - w->BorderRight;
		w->GZZHeight = w->Height - w->BorderTop  - w->BorderBottom;
	}

	if (nw.LeftEdge < 0 || nw.TopEdge < 0 ||
		nw.LeftEdge + w->Width > w->WScreen->Width ||
		nw.TopEdge + w->Height > w->WScreen->Height ||
		w->Width < 1 || w->Height < 1)
		goto failexit;

#ifdef __MORPHOS__
	IW(w)->parent = parentwin;
#endif

#ifndef __MORPHOS__
	if (NULL == parentwin)
	{
#endif
		w->LeftEdge    = nw.LeftEdge;
		w->TopEdge     = nw.TopEdge;
#ifndef __MORPHOS__
	}
	else
	{
		w->LeftEdge    = nw.LeftEdge + parentwin->LeftEdge;
		w->TopEdge     = nw.TopEdge  + parentwin->TopEdge;
	}
#endif

#ifndef __MORPHOS__
	w->RelLeftEdge = nw.LeftEdge;
	w->RelTopEdge  = nw.TopEdge;
#endif

	w->MinWidth  = (nw.MinWidth  != 0) ? nw.MinWidth  : w->Width;
	w->MinHeight = (nw.MinHeight != 0) ? nw.MinHeight : w->Height;
	w->MaxWidth  = (nw.MaxWidth  != 0) ? nw.MaxWidth  : w->Width;
	w->MaxHeight = (nw.MaxHeight != 0) ? nw.MaxHeight : w->Height;

	//jDc: tested behavior of intuition68k
	if ((UWORD)w->MaxWidth < w->Width) w->MaxWidth = w->Width;
	if ((UWORD)w->MaxHeight < w->Height) w->MaxHeight = w->Height;

	/* check if maxwidth/height is not bigger than screen */
	if (w->MaxWidth > w->WScreen->Width) w->MaxWidth = w->WScreen->Width;
	if (w->MaxHeight > w->WScreen->Height) w->MaxHeight = w->WScreen->Height;

	if (zoombox)
	{
		((struct IntWindow *)w)->ZipLeftEdge = zoombox->Left;
		((struct IntWindow *)w)->ZipTopEdge  = zoombox->Top;
		((struct IntWindow *)w)->ZipWidth    = zoombox->Width;
		((struct IntWindow *)w)->ZipHeight   = zoombox->Height;
	}
	else
	{
		//jDc: autodoc says it's maximum size
		((struct IntWindow *)w)->ZipWidth    = w->MaxWidth;
		((struct IntWindow *)w)->ZipHeight   = w->MaxHeight;

		((struct IntWindow *)w)->ZipLeftEdge = nw.LeftEdge;
		((struct IntWindow *)w)->ZipTopEdge  = nw.TopEdge;

		//in case we support offscreen layers we should clip to screen boundaries here!
		if (IW(w)->ZipLeftEdge + IW(w)->ZipWidth > w->WScreen->Width)
		{
			IW(w)->ZipLeftEdge = w->WScreen->Width - IW(w)->ZipWidth;
		}

		if (IW(w)->ZipTopEdge + IW(w)->ZipHeight > w->WScreen->Height)
		{
			IW(w)->ZipTopEdge  = w->WScreen->Height - IW(w)->ZipHeight;
		}
	}

	DEBUG_OPENWINDOW(dprintf("OpenWindow: zip %d %d %d %d\n",
				 ((struct IntWindow *)w)->ZipLeftEdge,
				 ((struct IntWindow *)w)->ZipTopEdge,
				 ((struct IntWindow *)w)->ZipWidth,
				 ((struct IntWindow *)w)->ZipHeight));


	IW(w)->InitialLeft = w->LeftEdge;
	IW(w)->InitialTop = w->TopEdge;
	IW(w)->InitialWidth = w->Width;
	IW(w)->InitialHeight = w->Height;

	IW(w)->mousequeue = mousequeue;
	IW(w)->repeatqueue = repeatqueue;

	/* Amiga and checkmark images for menus */

	IW(w)->Checkmark = Checkmark ? Checkmark :
			   ((struct IntScreen *)(w->WScreen))->DInfo.dri_CheckMark;

	IW(w)->AmigaKey  = AmigaKey  ? AmigaKey  :
			   ((struct IntScreen *)(w->WScreen))->DInfo.dri_AmigaKey;

#ifndef __MORPHOS__
	/* child support */
	if (NULL != parentwin)
	{
		if (parentwin->firstchild)
			parentwin->firstchild->prevchild = w;

		w->nextchild = parentwin->firstchild;
		parentwin->firstchild = w;
		w->parent = parentwin;
	}
#endif

#ifdef SKINS
	if (hasskininfo) ((struct IntWindow *)(w))->specialflags = SPFLAG_SKININFO;
	IW(w)->Application = (Object*)app;
#endif

#ifdef DAMAGECACHE
	if ((w->Flags & WFLG_SIMPLE_REFRESH) && (IS_DOCAREREFRESH(w)) && (!(w->Flags & WFLG_BORDERLESS))) IW(w)->trashregion = NewRegion();
#endif

	/* Help stuff */

	if (!have_helpgroup && helpgroupwindow)
	{
		if (IW(helpgroupwindow)->helpflags & HELPF_ISHELPGROUP)
		{
			helpgroup = IW(helpgroupwindow)->helpgroup;
			have_helpgroup = TRUE;
		}
	}

	if (have_helpgroup)
	{
		IW(w)->helpflags |= HELPF_ISHELPGROUP;
		IW(w)->helpgroup = helpgroup;
	}

	w->Title = nw.Title;

	/* as mentioned in the OpenScreen() autodoc/SA_SysFont settings the default
	window font will be either screenfont (sysfont=0) or defaultfont (sysfont=1) */
	w->IFont = SafeReopenFont(IntuitionBase, (IS(w->WScreen)->SysFont) ? &GfxBase->DefaultFont : &IS(w->WScreen)->DInfo.dri_Font);
	
	if (w->IFont == NULL) goto failexit;

	/* do the extrainit before opening a 1st window that could benefit from the extra components loaded */
	if ((usereffectbits != 0xFFFFFFFF) || (!(w->Flags & WFLG_BORDERLESS)))
		int_LayersExtraInit((struct Screen *)w->WScreen,IntuitionBase);

	/* jDc: intui68k waits before opening the window until
	** Move/SizeWindow actions are over (does it mean it's executed on
	** caller's context?).
	*/

#ifdef USEWINDOWLOCK
	if (!(FindTask(0) == ((struct IIHData *)IBase->InputHandler->is_Data)->InputDeviceTask))
	{
		LOCKWINDOW;
	}
#endif

	msg.window = w;
	msg.bitmap = nw.BitMap;
	//msg.backfillhook = backfillhook == LAYERS_BACKFILL ? &IW(w)->custombackfill : backfillhook;
	msg.backfillhook = backfillhook;
	msg.shape = shape;
	msg.parentlayer = parentl;
	msg.visible = windowvisible;
	msg.realtime = realtime;

	DoSyncAction((APTR)int_openwindow, &msg.msg, IntuitionBase);

#ifdef USEWINDOWLOCK
	if (!(FindTask(0) == ((struct IIHData *)IBase->InputHandler->is_Data)->InputDeviceTask))
	{
		UNLOCKWINDOW;
	}
#endif

	if (!msg.success)
		goto failexit;

	/* nlorentz: The driver has in some way or another allocated a rastport for us,
	   which now is ready for us to use. */

	driver_init_done = TRUE;
	rp = w->RPort;

	D(bug("called driver, rp=%p\n", rp));

	/* The window RastPort always gets the font from GfxBase->DefaultFont, which
	   is the system's default font. Usually topaz 8, but it can be changed with
	   the Fonts prefs program to another fixed-sized font. */

	SetFont (rp, w->IFont);

	D(bug("set fonts\n"));

#warning Remove workaround!
	/* lbischoff: The following 4 Setxxx lines are a workaround for the InitRastPort
	   problem (Bug #75 in docs/BUGS). They ensure that at least a window's rastport
	   is initialized correctly. Remove them if they are not needed any longer!
	*/
	SetABPenDrMd (rp, rp->FgPen, rp->BgPen, rp->DrawMode);
	SetWriteMask (rp, rp->Mask);
	D(bug("set pens\n"));


	/* Send all GA_RelSpecial BOOPSI gadgets in the list the GM_LAYOUT msg */
	/*DoGMLayout(w->FirstGadget, w, NULL, -1, TRUE, IntuitionBase);

	if (NULL != w->FirstGadget)
	    RefreshGadgets (w->FirstGadget, w, NULL);
	*/

	if (nw.FirstGadget)
	{
		AddGList(w, nw.FirstGadget, -1, -1, NULL);
	}

	if (screenTitle != NULL)
		SetWindowTitles (w, (UBYTE *)~0L, screenTitle);

	UpdateMouseCoords(w);

	if (do_setwindowpointer)
	{
		//jDc: main for () loop destroys original taglist pointer, we need to get
		//it once again here!
		tagList = (struct TagItem *)((struct ExtNewWindow *)newWindow)->Extension;
		SetWindowPointerA(w, tagList);
	} else {
		UpdateScreenPointerAsync(w->WScreen,IntuitionBase);
	}

	goto exit;

failexit:
	D(bug("fail\n"));

	if (w)
	{
		ModifyIDCMP (w, 0L);

		/* nlorentz: Freeing the rasport is now intui_CloseWindow()'s task.

		   if (rp)
		   {
		       FreeRastPort (rp);
		   }
		*/

		if (driver_init_done)
			intui_CloseWindow(w, IntuitionBase);

		if (w->IFont) CloseFont(w->IFont);

		#ifdef __MORPHOS__
		DisposeObject((APTR)w);
		#else
		FreeMem (w, sizeof(struct IntWindow));
		#endif

		w = NULL;
	}

	if (nw.Screen && (moreFlags & WMFLG_DO_UNLOCKPUBSCREEN))
	{
		UnlockPubScreen(NULL, nw.Screen);
	}

exit:

	DEBUG_OPENWINDOW(dprintf("OpenWindow: Return 0x%lx\n", w));

	ReturnPtr ("OpenWindow", struct Window *, w);
	AROS_LIBFUNC_EXIT
} /* OpenWindow */



/**********************************************************************************/

static VOID int_openwindow(struct OpenWindowActionMsg *msg,
						   struct IntuitionBase *IntuitionBase)
{
	struct Window * w = msg->window;
	struct BitMap * SuperBitMap = msg->bitmap;
	struct Hook   * backfillhook = msg->backfillhook;
#ifdef CreateLayerTagList
	struct Region * shape = msg->shape;
	struct Layer  * parent = msg->parentlayer;
	BOOL            visible = msg->visible;
#endif

#ifdef SKINS
	BOOL            installtransphook = FALSE;
	BOOL            installgzztransphook = FALSE;
	BOOL            notransphook = TRUE;
	struct windowclassprefs *wcprefs = NULL;
	BOOL            lidone = ATOMIC_BTST(&IS(w->WScreen)->DInfo.dri_Customize->LayersExtraInit,1);

#endif

	/* Create a layer for the window */
	LONG layerflags = 0;

	EnterFunc(bug("int_OpenWindow(w=%p)\n", w));

#ifdef SKINS
#if 0
	if (w->Flags & WFLG_ACTIVATE) QueryBlankerEvent(BLANKEREVENT_NEWWINDOW,(ULONG)IS(w->WScreen)->IMonitorNode,IntuitionBase);
#endif
#endif

	D(bug("screen: %p\n", w->WScreen));
	D(bug("bitmap: %p\n", w->WScreen->RastPort.BitMap));

	/* Set the layer's flags according to the flags of the
	** window
	*/

	/* refresh type - 68k intuition does this in a pretty funny way:
	** WFLG_OTHER_REFRESH (WFLG_SIMPLE_REFRESH | WFLG_SUPER_BITMAP) results
	** in a smart refresh window, even if the SuperBitMap was passed in!
	** The final w->Flags aren't changed in any way though. */
	if ((w->Flags & WFLG_REFRESHBITS) == WFLG_SIMPLE_REFRESH)
		layerflags |= LAYERSIMPLE;
	else
		if (((w->Flags & WFLG_REFRESHBITS) == WFLG_SUPER_BITMAP) && SuperBitMap)
			layerflags |= LAYERSUPER;
		else
			layerflags |= LAYERSMART;

	if (w->Flags & WFLG_BACKDROP)
		layerflags |= LAYERBACKDROP;

	if (!(w->Flags & WFLG_GIMMEZEROZERO) && (w->MoreFlags & WMFLG_FRONTWINDOW))
	{
		layerflags &= ~LAYERBACKDROP;
		layerflags |= LAYERONTOP;
	}

	D(bug("Window dims: (%d, %d, %d, %d)\n",
	      w->LeftEdge, w->TopEdge, w->Width, w->Height));

#ifdef SKINS
	//install transp layer hook!
	wcprefs = (struct windowclassprefs*)int_GetCustomPrefs(TYPE_WINDOWCLASS,&((struct IntScreen*)(w->WScreen))->DInfo,IntuitionBase);

	if (wcprefs->flags & WINDOWCLASS_PREFS_ROUNDEDEDGES && (w->Title || (w->Flags & (WFLG_DRAGBAR | WFLG_CLOSEGADGET | WFLG_DEPTHGADGET))))
	{
		if (w->Flags & WFLG_GIMMEZEROZERO)
		{
			IW(w)->gzztranspregion = NewRegion();
			if (IW(w)->gzztranspregion)
			{
				installgzztransphook = TRUE;
			}

			if (IW(w)->usertransphook || IW(w)->usertranspregion)
			{
				IW(w)->transpregion = NewRegion();
				if (IW(w)->transpregion)
				{
					installtransphook = TRUE;
					IW(w)->specialflags |= SPFLAG_TRANSPHOOK;
				}
			}
		} else {
			IW(w)->transpregion = NewRegion();
			if (IW(w)->transpregion)
			{
				installtransphook = TRUE; notransphook = FALSE;
				IW(w)->specialflags |= SPFLAG_TRANSPHOOK;
			}
		}
	} else {
		if (IW(w)->usertransphook)
		{
			IW(w)->transpregion = NewRegion();
			if (IW(w)->transpregion)
			{
				installtransphook = TRUE;
				IW(w)->specialflags |= SPFLAG_TRANSPHOOK;
			}
		} else if (IW(w)->usertranspregion)
		{
			IW(w)->transpregion = NewRegion();
			if (IW(w)->transpregion)
			{
				installtransphook = TRUE;
				IW(w)->specialflags |= SPFLAG_TRANSPHOOK;
			}
		}
	}
#endif

	LockLayers(&w->WScreen->LayerInfo);

	/* A GimmeZeroZero window??? */
	if (w->Flags & WFLG_GIMMEZEROZERO)
	{
		/*
		  A GimmeZeroZero window is to be created:
		    - the outer window will be a simple refresh layer
		    - the inner window will be a layer according to the flags
		  What is the size of the inner/outer window supposed to be???
		  I just make it that the outer window has the size of what is requested
		*/

#ifdef __MORPHOS__
		struct TagItem gzzlayertags[] =
		{
			{LA_BackfillHook, (IPTR)LAYERS_NOBACKFILL},
			{SuperBitMap ? LA_SuperBitMap : TAG_IGNORE, (IPTR)SuperBitMap},
			{installgzztransphook ? LA_TransHook : TAG_IGNORE,(IPTR)&((struct IntIntuitionBase *)(IntuitionBase))->transphook},
			{installgzztransphook ? LA_TransRegion : TAG_IGNORE, (IPTR)IW(w)->gzztranspregion},
			{LA_WindowPtr, (IPTR)w},
			{LA_Opacity, IW(w)->useropacity},
			{LA_OpacityMultiplier, (w->MoreFlags & WMFLG_LAYERMANAGER) ? 0 : 255},
			{LA_SourceAlpha,IW(w)->usersourcealpha},
			{LA_HitLevel,IW(w)->userhitlevel},
			{(IW(w)->EffectBits != 0) ? LA_EffectBits : TAG_IGNORE, IW(w)->EffectBits},
			{(IW(w)->EffectBits != 0) ? LA_GoBuffered : TAG_IGNORE, TRUE},
			{lidone ? LA_AlternateEffect : TAG_IGNORE,(w->Flags & WFLG_BORDERLESS) ? TRUE : FALSE},
			{TAG_DONE}
		};

		/* First create outer window */
		struct Layer * L = CreateUpfrontLayerTagList(
						        &w->WScreen->LayerInfo
						        , w->WScreen->RastPort.BitMap
						        , w->LeftEdge
						        , w->TopEdge
						        , w->LeftEdge + w->Width - 1
						        , w->TopEdge  + w->Height - 1
						        , LAYERSIMPLE | (layerflags & LAYERBACKDROP)
								, (struct TagItem *)&gzzlayertags);
#else

		/* First create outer window */
		struct Layer * L = CreateUpfrontHookLayer(
							       &w->WScreen->LayerInfo
							       , w->WScreen->RastPort.BitMap
							       , w->LeftEdge
							       , w->TopEdge
							       , w->LeftEdge + w->Width - 1
							       , w->TopEdge  + w->Height - 1
							       , LAYERSIMPLE | (layerflags & LAYERBACKDROP)
							       , LAYERS_NOBACKFILL
							       , NULL);

#endif
		/* Could the layer be created. Nothing bad happened so far, so simply leave */
		if (NULL == L)
		{
			msg->success = FALSE;
			UnlockLayers(&w->WScreen->LayerInfo);
			ReturnVoid("intui_OpenWindow(No GimmeZeroZero layer)");
		}

		D(bug("created outer GimmeZeroZero layer.\n"));

		/* install it as the BorderRPort */
		w->BorderRPort = L->rp;
		BLAYER(w) = L;

		/* This layer belongs to a window */
#ifndef __MORPHOS__
		L->Window = (APTR)w;
#endif

#ifdef __MORPHOS__
		{
			struct TagItem layertags[] =
			{
				{LA_BackfillHook, (IPTR)backfillhook},
				{SuperBitMap ? LA_SuperBitMap : TAG_IGNORE, (IPTR)SuperBitMap},
				{installtransphook ? LA_TransHook : TAG_IGNORE, notransphook ?
											(IPTR)&((struct IntIntuitionBase *)(IntuitionBase))->notransphook :
											(IPTR)&((struct IntIntuitionBase *)(IntuitionBase))->transphook},
				{installtransphook ? LA_TransRegion : TAG_IGNORE, (IPTR)IW(w)->transpregion},
				{LA_WindowPtr, (IPTR)w},
				{LA_Opacity, IW(w)->useropacity},
				{LA_OpacityMultiplier, (w->MoreFlags & WMFLG_LAYERMANAGER) ? 0 : 255},
				{LA_SourceAlpha,IW(w)->usersourcealpha},
				{LA_HitLevel,IW(w)->userhitlevel},
				{TAG_DONE}

			};
			/* Now comes the inner window */

			w->WLayer = CreateUpfrontLayerTagList(
					&w->WScreen->LayerInfo
					, w->WScreen->RastPort.BitMap
					, w->LeftEdge + w->BorderLeft
					, w->TopEdge  + w->BorderTop
					, w->LeftEdge + w->BorderLeft + w->GZZWidth - 1
					, w->TopEdge  + w->BorderTop + w->GZZHeight - 1
					, layerflags
					, (struct TagItem *)&layertags);
		}
#else
		/* Now comes the inner window */
		w->WLayer = CreateUpfrontHookLayer(
				&w->WScreen->LayerInfo
				, w->WScreen->RastPort.BitMap
				, w->LeftEdge + w->BorderLeft
				, w->TopEdge  + w->BorderTop
				, w->LeftEdge + w->BorderLeft + w->GZZWidth - 1
				, w->TopEdge  + w->BorderTop + w->GZZHeight - 1
				, layerflags
				, backfillhook
				, SuperBitMap);
#endif

		/* could this layer be created? If not then delete the outer window and exit */
		if (NULL == w->WLayer)
		{
			DeleteLayer(0, L);
			msg->success = FALSE;
			UnlockLayers(&w->WScreen->LayerInfo);
			ReturnVoid("intui_OpenWindow(No window layer)");
		}

		/* That should do it, I guess... */
	}
	else
	{
#ifdef __MORPHOS__
		struct TagItem layertags[] =
		{
			{LA_BackfillHook, (IPTR)backfillhook},
			{SuperBitMap ? LA_SuperBitMap : TAG_IGNORE, (IPTR)SuperBitMap},
			{installtransphook ? LA_TransHook : TAG_IGNORE, notransphook ? (IPTR)&((struct IntIntuitionBase *)(IntuitionBase))->notransphook : (IPTR)&((struct IntIntuitionBase *)(IntuitionBase))->transphook},
			{installtransphook ? LA_TransRegion : TAG_IGNORE, (IPTR)IW(w)->transpregion},
			{LA_WindowPtr, (IPTR)w},
			{LA_Opacity, IW(w)->useropacity},
			{LA_OpacityMultiplier, (w->MoreFlags & WMFLG_LAYERMANAGER) ? 0 : 255},
			{LA_SourceAlpha,IW(w)->usersourcealpha},
			{LA_HitLevel,IW(w)->userhitlevel},
			{(IW(w)->EffectBits != 0) ? LA_EffectBits : TAG_IGNORE, IW(w)->EffectBits},
			{(IW(w)->EffectBits != 0) ? LA_GoBuffered : TAG_IGNORE, TRUE},
			{lidone ? LA_AlternateEffect : TAG_IGNORE,(w->Flags & WFLG_BORDERLESS) ? TRUE : FALSE},
			{LA_Realtime, msg->realtime},
			{TAG_DONE}
		};

		// in case lidone is false it means no window with a possible shadow or transp was opened yet
		// it also means that THIS window doesn't need transparency, therefore it's not really important
		// if we set the LA_AlternateEffect here for borderless windows (as long as windowclass will do that
		// if a shadow is requested at a later time!

		D(dprintf("CreateUpfontLayerTagList(taglist 0x%lx)\n",&layertags));

		w->WLayer = CreateUpfrontLayerTagList(		&w->WScreen->LayerInfo,
								w->WScreen->RastPort.BitMap,
								w->LeftEdge,
								w->TopEdge,
								w->LeftEdge + w->Width - 1,
								w->TopEdge  + w->Height - 1,
								layerflags,
								(struct TagItem *)&layertags);
#else

#ifdef CreateLayerTagList
		struct TagItem win_tags[] =
		    {
			    {LA_Hook     , (IPTR)backfillhook},
			    {LA_Priority	    , (layerflags & LAYERBACKDROP) ? BACKDROPPRIORITY : UPFRONTPRIORITY },
			    {LA_Shape	    , (IPTR)shape							},
			    {LA_SuperBitMap   , (IPTR)SuperBitMap							},
			    {LA_ChildOf       , (IPTR)parent							},
			    {LA_Visible       , (ULONG)visible                                                    },
			    {TAG_DONE										}
		    };

		w->WLayer = CreateLayerTagList(&w->WScreen->LayerInfo,
					       w->WScreen->RastPort.BitMap,
					       w->RelLeftEdge,
					       w->RelTopEdge,
					       w->RelLeftEdge + w->Width - 1,
					       w->RelTopEdge + w->Height - 1,
					       layerflags,
					       win_tags);

#else
		w->WLayer = CreateUpfrontHookLayer(
				&w->WScreen->LayerInfo
				, w->WScreen->RastPort.BitMap
				, w->RelLeftEdge
				, w->RelTopEdge
				, w->LeftEdge + w->Width - 1
				, w->TopEdge  + w->Height - 1
				, layerflags
				, backfillhook
				, SuperBitMap);
#endif
#endif

		/* Install the BorderRPort here! see GZZ window above  */
		if (NULL != w->WLayer)
		{
			/*
			   I am installing a totally new RastPort here so window and frame can
			   have different fonts etc.
			*/
			w->BorderRPort = AllocMem(sizeof(struct RastPort), MEMF_ANY);
			if (w->BorderRPort)
			{
				InitRastPort(w->BorderRPort);
				w->BorderRPort->Layer  = w->WLayer;
				w->BorderRPort->BitMap = w->WLayer->rp->BitMap;
			}
			else
			{
				/* no memory for RastPort! Simply close the window */
				intui_CloseWindow(w, IntuitionBase);
				msg->success = FALSE;
				UnlockLayers(&w->WScreen->LayerInfo);
				ReturnVoid("intui_OpenWindow(No BorderRPort)");
			}
		}
	}

	D(bug("Layer created: %p\n", w->WLayer));
	D(bug("Window created: %p\n", w));

	/* common code for GZZ and regular windows */

	if (w->WLayer)
	{

		if ((layerflags & LAYERBACKDROP) && (w->WScreen->Flags & SHOWTITLE))
		{
#ifdef __MORPHOS__
			struct Layer *blayer;

			/* make sure the screen titlebar is in front of all user created */
			/* backdrop windows */

			blayer = w->WScreen->BarLayer;

#ifdef SKINS
			if (IS(w->WScreen)->SpecialFlags & (SF_InvisibleBar|SF_AppearingBar)) blayer = 0;
#endif

			if (blayer) MoveLayerInFrontOf(blayer,w->WLayer);

			D(bug("move screen bar layer in front of window backdrop layer\n"));
#else
			/* backdrop window was created over screen barlayer, but it must be
			under the screen barlayer if screen has flag SHOWTITLE set */

			Forbid();
			w->WScreen->Flags &= ~SHOWTITLE;
			Permit();

			ShowTitle(w->WScreen, TRUE);
#endif
		}
		else
		{
			if ((IS(w->WScreen)->SpecialFlags & SF_AppearingBar) && !MENUS_ACTIVE)
			{
				UpfrontLayer(0,w->WScreen->BarLayer);
			}
		}

		/* Layer gets pointer to the window */
		WLAYER(w) = w->WLayer;

		/* set the border separation, if required ! */
#ifdef SKINS
		if (wcprefs->flags & WINDOWCLASS_PREFS_TITLEBLEND)
		{
			if ((!(w->Flags & WFLG_BORDERLESS)) && w->Title)
			{
				struct Rectangle borderrect = {0,w->BorderTop,0,0};
				LayersControlTags(&w->WScreen->LayerInfo,
					LT_Layer,(ULONG)BLAYER(w) ? (ULONG)BLAYER(w) : (ULONG)WLAYER(w),
					LT_BorderSeparation,(ULONG)&borderrect,
					LT_BorderSourceAlpha, TRUE,
					TAG_DONE);
			}
		}
#endif
		
		CheckLayers(w->WScreen,IntuitionBase);

		UnlockLayers(&w->WScreen->LayerInfo);

		if (WMFLG_LAYERMANAGER == (w->MoreFlags & WMFLG_LAYERMANAGER))
		{
			windowLayersOpened(w, WFLG_ACTIVATE == (w->Flags & WFLG_ACTIVATE), IntuitionBase);
		}

#ifndef __MORPHOS__
		w->WLayer->Window = (APTR)w;
#endif
		/* Window needs a rastport */
		w->RPort = w->WLayer->rp;

		/* installation of the correct BorderRPort already happened above !! */

		if (CreateWinSysGadgets(w, IntuitionBase))
		{
			struct Window *parent, *descendant_of_parent;
			LONG lock = 0;
			
			if (!ILOCKCHECK(((struct IntuiActionMsg *)msg))) lock = LockIBase (0);

			/* insert new window into parent/descendant list
			**
			** before: parent win xyz
			**            |
			**            |
			**            |
			**        descendant win abc
			**
			** after:  parent win xyz
			**              \
			**		 \
			**	       newwindow w
			**		/
			**	       /
			**	      /
			**	 descendant win abc
			*/

#if 1
			parent = IntuitionBase->ActiveWindow;
			if (!parent) parent = w->WScreen->FirstWindow;
			if (parent)
			{
				descendant_of_parent = parent->Descendant;
				parent->Descendant = w;
				if (descendant_of_parent) descendant_of_parent->Parent = w;
			}
			else
			{
				descendant_of_parent = NULL;
			}

			w->Descendant = descendant_of_parent;
			w->Parent = parent;
#endif
			w->NextWindow = w->WScreen->FirstWindow;
			w->WScreen->FirstWindow = w;

			w->WindowPort = IBase->IntuiReplyPort;

			if (!ILOCKCHECK(((struct IntuiActionMsg *)msg))) UnlockIBase (lock);

			if ((w->Flags & WFLG_ACTIVATE) && (!(w->MoreFlags & WMFLG_TOOLBARWINDOW)))
			{
				struct IIHData  *iihdata = (struct IIHData *)IBase->InputHandler->is_Data;
				struct ActivateWindowActionMsg awmsg;

				awmsg.msg.task = msg->msg.task;
				awmsg.window = w;
				awmsg.useLayerManager = FALSE;
				awmsg.allowAnimations = FALSE;

				if (iihdata->ActiveGadget && (((struct IntuiActionMsg *)msg)->task != FindTask(0)) &&
					((iihdata->ActiveGadget != iihdata->MasterDragGadget) && (iihdata->ActiveGadget != iihdata->MasterSizeGadget)))
				IHInactiveGadget(iihdata);

				int_activatewindow(&awmsg,IntuitionBase);
			}
			else
			{
				RefreshWindowFrame(w);
				CheckMouseZone(FALSE,IntuitionBase);
				UpdateScreenPointer(w->WScreen,IntuitionBase);
			}

			msg->success = TRUE;
			ReturnVoid("int_openwindow");
		}

		CloseWindow(w);
		//int_closewindow(w, IntuitionBase);

	} /* if (layer created) */

	UnlockLayers(&w->WScreen->LayerInfo);


	D(bug("int_openwindow(General failure)"));

	msg->success = FALSE;
	ReturnVoid("int_openwindow");
}
